@@ -19,16 +19,22 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
-
 import copy
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Union
 
 import numpy as np
-
-from scipy.stats import truncnorm, norm
+from scipy import stats
 from scipy.sparse import diags
 
 from . import site
 
+# Used to define the random state. A specific state can be set with:
+# random_state.set_seed(42)
+random_state = np.random.RandomState()
+
 
 class TruncatedNorm:
     """Truncated normal random number generator.
@@ -55,7 +61,7 @@ class TruncatedNorm:
         # deviations, the input standard deviation must be increased to
         # 1.136847 to maintain a unit standard deviation for the random
         # samples.
-        self._scale = 1 / np.sqrt(truncnorm.stats(-value, value, moments='v'))
+        self._scale = 1 / np.sqrt(stats.truncnorm.stats(-value, value, moments="v"))
 
     @property
     def scale(self):
@@ -81,12 +87,13 @@ class TruncatedNorm:
         rvs : ndarray or scalar
             Random variates of given `size`.
         """
-        return truncnorm.rvs(
-            -self.limit, self.limit, scale=self._scale, size=size)
+        return stats.truncnorm.rvs(
+            -self.limit, self.limit, scale=self._scale, size=size
+        )
 
     def correlated(self, correl):
         # Acceptance proportion
-        accept = np.diff(norm.cdf([-self.limit, self.limit]))[0]
+        accept = np.diff(stats.norm.cdf([-self.limit, self.limit]))[0]
         # The expected number of tries required
         expected = np.ceil(1 / accept).astype(int)
 
@@ -95,7 +102,8 @@ class TruncatedNorm:
             # specified standard deviation. Use twice the expected since
             # this calculation is fast and we don't want to loop.
             randvar = np.random.multivariate_normal(
-                [0, 0], [[1, correl], [correl, 1]], size=(2 * expected))
+                [0, 0], [[1, correl], [correl, 1]], size=(2 * expected)
+            )
             valid = np.all(np.abs(randvar) < self.limit, axis=1)
             if np.any(valid):
                 # Return the first valid value
@@ -104,11 +112,12 @@ class TruncatedNorm:
 
 # Random number generator used for all random number. Limited to +/- 2,
 # and the standard deviation is scaled to maintain the standard deviation
+# FIXME
 randnorm = TruncatedNorm(2)
 
 
 class ToroThicknessVariation(object):
-    """ Toro (1995) [T95]_ thickness variation model.
+    """Toro (1995) [T95]_ thickness variation model.
 
     The recommended values are provided as defaults to this model.
 
@@ -178,14 +187,20 @@ class ToroThicknessVariation(object):
             total += np.random.exponential(1.0)
 
             # Convert between x and depth using the inverse of \Lambda(t)
-            depth = np.power(
-                (self.c_2 * total) / self.c_3 + total / self.c_3 + np.power(
-                    self.c_1, self.c_2 + 1), 1 / (self.c_2 + 1)) - self.c_1
+            depth = (
+                np.power(
+                    (self.c_2 * total) / self.c_3
+                    + total / self.c_3
+                    + np.power(self.c_1, self.c_2 + 1),
+                    1 / (self.c_2 + 1),
+                )
+                - self.c_1
+            )
 
             thickness = depth - depth_prev
 
             if depth > depth_total:
-                thickness = (depth_total - depth_prev)
+                thickness = depth_total - depth_prev
                 depth = depth_prev + thickness
 
             depth_mid = (depth_prev + depth) / 2
@@ -212,8 +227,7 @@ class ToroThicknessVariation(object):
             # Locate the proper layer and add it to the model
             for l in profile:
                 if l.depth < depth_mid <= l.depth_base:
-                    layers.append(
-                        site.Layer(l.soil_type, thick, l.initial_shear_vel))
+                    layers.append(site.Layer(l.soil_type, thick, l.initial_shear_vel))
                     break
             else:
                 raise LookupError
@@ -226,10 +240,44 @@ class ToroThicknessVariation(object):
         return varied
 
 
+class BedrockDepthVariation(object):
+    def __init__(
+        self,
+        dist: stats.rv_continuous,
+        discretize_kwds: Optional[Dict[str, float]] = None,
+    ):
+        self._dist = dist
+        self._discretize_kwds = discretize_kwds or dict(max_freq=50, wave_frac=2)
+
+    def __call__(self, profile: site.Profile) -> site.Profile:
+        varied_depth = self._dist.rvs()
+
+        index, depth_within = profile.lookup_depth(varied_depth)
+
+        _profile = site.Profile([layer for layer in profile[: (index + 1)]])
+        _profile[-1].thickness = depth_within
+
+        return _profile.auto_discretize(**self._discretize_kwds)
+
+
+class LayerThicknessVariation(object):
+    def __init__(
+        self,
+        models: Union[List[stats.rv_continuous], Dict[int, stats.rv_continuous]],
+        discretize_kwds: Optional[Dict[str, float]] = None,
+    ) -> None:
+
+        self._models = models
+        self._discretize_kwds = discretize_kwds
+
+
 class VelocityVariation(object):
     """Abstract model for varying the velocity."""
 
-    def __call__(self, profile):
+    def __init__(self, vary_bedrock=False):
+        self._vary_bedrock = vary_bedrock
+
+    def __call__(self, profile: site.Profile) -> site.Profile:
         """Calculate a varied shear-wave velocity profile.
 
         Parameters
@@ -243,22 +291,26 @@ class VelocityVariation(object):
             Varied site profile.
         """
 
-        mean = np.log([l.initial_shear_vel for l in profile])
+        mean = np.log(profile.initial_shear_vel)
         covar = self._calc_covar_matrix(profile)
 
-        lnvel_rand = np.random.multivariate_normal(
-            mean, covar, check_valid='ignore')
+        ln_vel_rand = np.random.multivariate_normal(mean, covar, check_valid="ignore")
 
         # Limits based on the number of standard deviations
         offset = randnorm.limit * np.sqrt(np.diag(covar))
-        lnvel = np.clip(lnvel_rand, mean - offset, mean + offset)
-
-        vel = np.exp(lnvel)
+        ln_vel = np.clip(ln_vel_rand, mean - offset, mean + offset)
+        vel = np.exp(ln_vel)
 
+        # Create the new layers
+        end = None if self.vary_bedrock else -1
+        # FIXME: Can we use .copy()?
         layers = [
             site.Layer(l.soil_type, l.thickness, vel[i])
-            for i, l in enumerate(profile)
+            for i, l in enumerate(profile[:end])
         ]
+        if not self.vary_bedrock:
+            layers.append(profile[-1])
+
         varied = site.Profile(layers, profile.wt_depth)
         return varied
 
@@ -280,7 +332,7 @@ class VelocityVariation(object):
         # Modify the standard deviation by the truncated norm scale
         std *= randnorm.scale
 
-        var = std ** 2
+        var = std**2
         covar = corr * std[:-1] * std[1:]
 
         # Main diagonal is the variance
@@ -318,9 +370,13 @@ class VelocityVariation(object):
         """
         raise NotImplementedError
 
+    @property
+    def vary_bedrock(self):
+        return self._vary_bedrock
+
 
 class ToroVelocityVariation(VelocityVariation):
-    """ Toro (1995) [T95] velocity variation model.
+    """Toro (1995) [T95] velocity variation model.
 
     Default values can be selected with :meth:`.generic_model`.
 
@@ -341,74 +397,85 @@ class ToroVelocityVariation(VelocityVariation):
     """
 
     PARAMS = {
-        'Geomatrix AB': {
-            'ln_std': 0.46,
-            'rho_0': 0.96,
-            'delta': 13.1,
-            'rho_200': 0.96,
-            'h_0': 0.0,
-            'b': 0.095,
+        "Geomatrix AB": {
+            "ln_std": 0.46,
+            "rho_0": 0.96,
+            "delta": 13.1,
+            "rho_200": 0.96,
+            "h_0": 0.0,
+            "b": 0.095,
         },
-        'Geomatrix CD': {
-            'ln_std': 0.38,
-            'rho_0': 0.99,
-            'delta': 8.0,
-            'rho_200': 1.00,
-            'h_0': 0.0,
-            'b': 0.160,
+        "Geomatrix CD": {
+            "ln_std": 0.38,
+            "rho_0": 0.99,
+            "delta": 8.0,
+            "rho_200": 1.00,
+            "h_0": 0.0,
+            "b": 0.160,
         },
-        'USGS AB': {
-            'ln_std': 0.35,
-            'rho_0': 0.95,
-            'delta': 4.2,
-            'rho_200': 1.00,
-            'h_0': 0.0,
-            'b': 0.138,
+        "USGS AB": {
+            "ln_std": 0.35,
+            "rho_0": 0.95,
+            "delta": 4.2,
+            "rho_200": 1.00,
+            "h_0": 0.0,
+            "b": 0.138,
         },
-        'USGS CD': {
-            'ln_std': 0.36,
-            'rho_0': 0.99,
-            'delta': 3.9,
-            'rho_200': 1.00,
-            'h_0': 0.0,
-            'b': 0.293,
+        "USGS CD": {
+            "ln_std": 0.36,
+            "rho_0": 0.99,
+            "delta": 3.9,
+            "rho_200": 1.00,
+            "h_0": 0.0,
+            "b": 0.293,
         },
-        'USGS A': {
-            'ln_std': 0.36,
-            'rho_0': 0.95,
-            'delta': 3.4,
-            'rho_200': 0.42,
-            'h_0': 0.0,
-            'b': 0.063,
+        "USGS A": {
+            "ln_std": 0.36,
+            "rho_0": 0.95,
+            "delta": 3.4,
+            "rho_200": 0.42,
+            "h_0": 0.0,
+            "b": 0.063,
         },
-        'USGS B': {
-            'ln_std': 0.27,
-            'rho_0': 0.97,
-            'delta': 3.8,
-            'rho_200': 1.00,
-            'h_0': 0.0,
-            'b': 0.293,
+        "USGS B": {
+            "ln_std": 0.27,
+            "rho_0": 0.97,
+            "delta": 3.8,
+            "rho_200": 1.00,
+            "h_0": 0.0,
+            "b": 0.293,
         },
-        'USGS C': {
-            'ln_std': 0.31,
-            'rho_0': 0.99,
-            'delta': 3.9,
-            'rho_200': 0.98,
-            'h_0': 0.0,
-            'b': 0.344,
+        "USGS C": {
+            "ln_std": 0.31,
+            "rho_0": 0.99,
+            "delta": 3.9,
+            "rho_200": 0.98,
+            "h_0": 0.0,
+            "b": 0.344,
         },
-        'USGS D': {
-            'ln_std': 0.37,
-            'rho_0': 0.00,
-            'delta': 5.0,
-            'rho_200': 0.50,
-            'h_0': 0.0,
-            'b': 0.744,
+        "USGS D": {
+            "ln_std": 0.37,
+            "rho_0": 0.00,
+            "delta": 5.0,
+            "rho_200": 0.50,
+            "h_0": 0.0,
+            "b": 0.744,
         },
     }
 
-    def __init__(self, ln_std, rho_0, delta, rho_200, h_0, b):
+    def __init__(
+        self,
+        ln_std: float,
+        rho_0: float,
+        delta: float,
+        rho_200: float,
+        h_0: float,
+        b: float,
+        vary_bedrock: bool = False,
+    ):
         """Initialize the model."""
+        super().__init__(vary_bedrock=vary_bedrock)
+
         self._ln_std = ln_std
         self._rho_0 = rho_0
         self._delta = delta
@@ -416,7 +483,7 @@ class ToroVelocityVariation(VelocityVariation):
         self._h_0 = h_0
         self._b = b
 
-    def _calc_corr(self, profile):
+    def _calc_corr(self, profile: site.Profile) -> np.ndarray:
         """Compute the adjacent-layer correlations
 
         Parameters
@@ -429,13 +496,14 @@ class ToroVelocityVariation(VelocityVariation):
         corr : :class:`numpy.array`
             Adjacent-layer correlations
         """
-        depth = np.array([l.depth_mid for l in profile[:-1]])
+        depth = profile.depth_mid[:-1]
         thick = np.diff(depth)
         depth = depth[1:]
 
         # Depth dependent correlation
-        corr_depth = (self.rho_200 * np.power(
-            (depth + self.rho_0) / (200 + self.rho_0), self.b))
+        corr_depth = self.rho_200 * np.power(
+            (depth + self.rho_0) / (200 + self.rho_0), self.b
+        )
         corr_depth[depth > 200] = self.rho_200
 
         # Thickness dependent correlation
@@ -521,11 +589,121 @@ class ToroVelocityVariation(VelocityVariation):
         return cls(**p)
 
 
+class DepthDependToroVelVariation(ToroVelocityVariation):
+    """Toro (1995) [T95] velocity variation model modified for a depth
+    dependent standard deviation that can be overridden by the soil_type name.
+
+    Default values can be selected with :meth:`.generic_model`.
+
+    Parameters
+    ----------
+    ln_std_map: dict[str, float], optional
+        Mapping between the soil_type and the defined ln_std. Default is *None*.
+    depth: array_like, optional
+        Depths defining the standard deviation model. Default is [0, 15]
+        following the SPID model.
+    ln_std: array_like, optional
+        :math:`\sigma_{ln}` model parameter. Default is [0.25, 0.15]
+        following the SPID model.
+    rho_0: float, optional
+        :math:`ρ_0` model parameter.
+    delta: float, optional
+        :math:`\Delta` model parameter.
+    rho_200: float, optional
+        :math:`ρ_200` model parameter.
+    h_0: float, optional
+        :math:`h_0` model parameter.
+    b: float, optional
+        :math:`b` model parameter.
+    """
+
+    def __init__(
+        self,
+        ln_std_map,
+        depth,
+        ln_std,
+        rho_0,
+        delta,
+        rho_200,
+        h_0,
+        b,
+        vary_bedrock=False,
+    ):
+        """Initialize the model."""
+        super().__init__(
+            depth, ln_std, rho_0, delta, rho_200, h_0, b, vary_bedrock=vary_bedrock
+        )
+        self._ln_std_map = ln_std_map or dict()
+
+    def _calc_ln_std(self, profile):
+        # Depth based values
+        ln_std = super()._calc_ln_std(profile)
+
+        # Update based on soil_type name
+        for name, value in self._ln_std_map.items():
+            for i, l in enumerate(profile):
+                if name in l.soil_type.name:
+                    ln_std[i] = value
+
+        return ln_std
+
+    @classmethod
+    def generic_model(cls, site_class, ln_std_map=None, **kwds):
+        """Use generic model parameters based on site class.
+
+        Parameters
+        ----------
+        site_class: str
+            Site classification. Possible options are:
+             * Geomatrix AB
+             * Geomatrix CD
+             * USGS AB
+             * USGS CD
+             * USGS A
+             * USGS B
+             * USGS C
+             * USGS D
+
+            See the report for definitions of the Geomatrix site
+            classication. USGS site classification is based on :math:`V_{s30}`:
+
+            =========== =====================
+            Site Class  :math:`V_{s30}` (m/s)
+            =========== =====================
+            A           >750 m/s
+            B           360 to 750 m/s
+            C           180 to 360 m/s
+            D           <180 m/s
+            =========== =====================
+
+        ln_std_map: dict[str, float], optional
+            Mapping between the soil_type and the defined ln_std. Default is *None*.
+
+        Returns
+        -------
+        DepthAndSoilTypeDependToroVelVariation
+            Initialized :class:`DepthDependToroVelVariation` with generic parameters.
+        """
+        p = dict(cls.PARAMS[site_class])
+        p.update(kwds)
+
+        if "depth" not in kwds:
+            p["depth"] = [0, 15]
+            p["ln_std"] = [0.25, 0.15]
+
+        p["ln_std_map"] = ln_std_map
+        return cls(**p)
+
+
 class SoilTypeVariation(object):
-    def __init__(self,
-                 correlation,
-                 limits_mod_reduc=[0.05, 1],
-                 limits_damping=[0, 0.15]):
+    def __init__(
+        self,
+        correlation,
+        limits_mod_reduc=[0.05, 1],
+        limits_damping=[0, 0.15],
+        vary_bedrock=False,
+    ):
+        self._vary_bedrock = vary_bedrock
         self._correlation = correlation
         self._limits_mod_reduc = list(limits_mod_reduc)
         self._limits_damping = list(limits_damping)
@@ -543,19 +721,21 @@ class SoilTypeVariation(object):
         # A pair of correlated random variables
         randvar = randnorm.correlated(self.correlation)
 
-        varied_mod_reduc, varied_damping = self._get_varied(randvar, mod_reduc,
-                                                            damping)
+        varied_mod_reduc, varied_damping = self._get_varied(randvar, mod_reduc, damping)
 
         # Clip the values to the specified min/max
-        varied_mod_reduc = np.clip(varied_mod_reduc, self.limits_mod_reduc[0],
-                                   self.limits_mod_reduc[1])
-        varied_damping = np.clip(varied_damping, self.limits_damping[0],
-                                 self.limits_damping[1])
+        varied_mod_reduc = np.clip(
+            varied_mod_reduc, self.limits_mod_reduc[0], self.limits_mod_reduc[1]
+        )
+        varied_damping = np.clip(
+            varied_damping, self.limits_damping[0], self.limits_damping[1]
+        )
 
         # Set the values
         realization = copy.deepcopy(soil_type)
-        for attr_name, values in zip(['mod_reduc', 'damping'],
-                                     [varied_mod_reduc, varied_damping]):
+        for attr_name, values in zip(
+            ["mod_reduc", "damping"], [varied_mod_reduc, varied_damping]
+        ):
             try:
                 getattr(realization, attr_name).values = values
             except AttributeError:
@@ -577,6 +757,10 @@ class SoilTypeVariation(object):
     def limits_mod_reduc(self):
         return self._limits_mod_reduc
 
+    @property
+    def vary_bedrock(self):
+        return self._vary_bedrock
+
 
 class DarendeliVariation(SoilTypeVariation):
     def _get_varied(self, randvar, mod_reduc, damping):
@@ -607,8 +791,9 @@ class DarendeliVariation(SoilTypeVariation):
             Standard deviation.
         """
         mod_reduc = np.asarray(mod_reduc).astype(float)
-        std = (np.exp(-4.23) + np.sqrt(0.25 / np.exp(3.62) - (mod_reduc - 0.5)
-                                       ** 2 / np.exp(3.62)))
+        std = np.exp(-4.23) + np.sqrt(
+            0.25 / np.exp(3.62) - (mod_reduc - 0.5) ** 2 / np.exp(3.62)
+        )
         return std
 
     @staticmethod
@@ -628,7 +813,7 @@ class DarendeliVariation(SoilTypeVariation):
             Standard deviation.
         """
         damping = np.asarray(damping).astype(float)
-        std = (np.exp(-5) + np.exp(-0.25) * np.sqrt(100 * damping)) / 100.
+        std = (np.exp(-5) + np.exp(-0.25) * np.sqrt(100 * damping)) / 100.0
         return std
 
 
@@ -636,12 +821,14 @@ class SpidVariation(SoilTypeVariation):
     """Variation defined by the EPRI SPID (2013) and documented in
     PNNL (2014)."""
 
-    def __init__(self,
-                 correlation,
-                 limits_mod_reduc=[0, 1],
-                 limits_damping=[0, 0.15],
-                 std_mod_reduc=0.15,
-                 std_damping=0.0030):
+    def __init__(
+        self,
+        correlation,
+        limits_mod_reduc=[0, 1],
+        limits_damping=[0, 0.15],
+        std_mod_reduc=0.15,
+        std_damping=0.30,
+    ):
         super().__init__(correlation, limits_mod_reduc, limits_damping)
         self._std_mod_reduc = std_mod_reduc
         self._std_damping = std_damping
@@ -649,7 +836,8 @@ class SpidVariation(SoilTypeVariation):
     def _get_varied(self, randvar, mod_reduc, damping):
         # Vary the G/Gmax in transformed space.
         # Equation 9.43 of PNNL (2014)
-        f_mean = (mod_reduc / (1 - mod_reduc))
+        # Here epsilon is added so that the denomiator doesn't go to zero.
+        f_mean = mod_reduc / (1 - mod_reduc + np.finfo(float).eps)
         # Instead of constraining the standard deviation at a specific
         # strain, then standard deviation is constrained at G/Gmax of 0.5.
         # This is modified from Equation 9.44 of PNNL (2014).
@@ -659,8 +847,7 @@ class SpidVariation(SoilTypeVariation):
         varied_mod_reduc = f_real / (1 + f_real)
 
         # Simple log distribution
-        varied_damping = \
-            np.exp(randvar[1] * self.std_damping) * damping
+        varied_damping = np.exp(randvar[1] * self.std_damping) * damping
 
         return varied_mod_reduc, varied_damping
 
@@ -673,11 +860,13 @@ class SpidVariation(SoilTypeVariation):
         return self._std_mod_reduc
 
 
-def iter_varied_profiles(profile,
-                         count,
-                         var_thickness=None,
-                         var_velocity=None,
-                         var_soiltypes=None):
+def iter_varied_profiles(
+    profile,
+    count,
+    var_thickness: ToroThicknessVariation = None,
+    var_velocity: VelocityVariation = None,
+    var_soiltypes: SoilTypeVariation = None,
+):
     for i in range(count):
         # Copy the profile to form the realization
         p = profile
@@ -692,10 +881,15 @@ def iter_varied_profiles(profile,
             # Map of varied soil types
             varied = {str(st): var_soiltypes(st) for st in p.iter_soil_types()}
             # Create new layers
+            end = None if var_soiltypes.vary_bedrock else -1
             layers = [
-                site.Layer(varied[str(l.soil_type)], l.thickness,
-                           l.initial_shear_vel) for l in p
+                site.Layer(varied[str(l.soil_type)], l.thickness, l.initial_shear_vel)
+                for l in p[:end]
             ]
+            # Add the unrandomized bedrock
+            if not var_soiltypes.vary_bedrock:
+                layers.append(p[-1])
+
             # Create a new profile
             p = site.Profile(layers, p.wt_depth)
 
